#!/usr/bin/env python3
"""
File Integrity Monitor
Advanced file system monitoring and integrity verification tool.
Parameters:
    - paths: List of paths to monitor
    - mode: Monitoring mode (realtime, periodic)
    - interval: Check interval for periodic mode
Requires: watchdog, cryptography packages
"""
import argparse
import logging
from datetime import datetime
import json
import threading
import queue
from collections import defaultdict
import os
import time
import hashlib
import sqlite3
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
import magic
import stat
import platform
import psutil
from pathlib import Path

class FileIntegrityMonitor:
    def __init__(self, paths, mode='realtime', interval=60):
        self.paths = [Path(p).resolve() for p in paths]
        self.mode = mode.lower()
        self.interval = interval
        self.db_path = 'file_integrity.db'
        self.event_queue = queue.Queue()
        self.stop_event = threading.Event()
        self.setup_logging()
        self.initialize_database()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def initialize_database(self):
        """Initialize SQLite database for file integrity information"""
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute('''
                    CREATE TABLE IF NOT EXISTS files (
                        path TEXT PRIMARY KEY,
                        hash TEXT,
                        size INTEGER,
                        permissions TEXT,
                        owner TEXT,
                        last_modified TEXT,
                        last_checked TEXT
                    )
                ''')
                
                cursor.execute('''
                    CREATE TABLE IF NOT EXISTS events (
                        timestamp TEXT,
                        event_type TEXT,
                        path TEXT,
                        details TEXT
                    )
                ''')
                conn.commit()
        except Exception as e:
            self.logger.error(f"Error initializing database: {str(e)}")

    def calculate_file_hash(self, file_path):
        """Calculate SHA-256 hash of file"""
        try:
            sha256_hash = hashlib.sha256()
            with open(file_path, "rb") as f:
                for byte_block in iter(lambda: f.read(4096), b""):
                    sha256_hash.update(byte_block)
            return sha256_hash.hexdigest()
        except Exception as e:
            self.logger.error(f"Error calculating hash for {file_path}: {str(e)}")
            return None

    def get_file_info(self, file_path):
        """Get detailed file information"""
        try:
            stat_info = os.stat(file_path)
            return {
                'path': str(file_path),
                'hash': self.calculate_file_hash(file_path),
                'size': stat_info.st_size,
                'permissions': oct(stat.S_IMODE(stat_info.st_mode)),
                'owner': stat_info.st_uid,
                'last_modified': datetime.fromtimestamp(
                    stat_info.st_mtime
                ).isoformat(),
                'last_checked': datetime.now().isoformat()
            }
        except Exception as e:
            self.logger.error(
                f"Error getting file info for {file_path}: {str(e)}"
            )
            return None

    def store_file_info(self, file_info):
        """Store file information in database"""
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute('''
                    INSERT OR REPLACE INTO files
                    (path, hash, size, permissions, owner,
                     last_modified, last_checked)
                    VALUES (?, ?, ?, ?, ?, ?, ?)
                ''', (
                    file_info['path'],
                    file_info['hash'],
                    file_info['size'],
                    file_info['permissions'],
                    file_info['owner'],
                    file_info['last_modified'],
                    file_info['last_checked']
                ))
                conn.commit()
        except Exception as e:
            self.logger.error(
                f"Error storing file info: {str(e)}"
            )

    def log_event(self, event_type, path, details):
        """Log file system event to database"""
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute('''
                    INSERT INTO events
                    (timestamp, event_type, path, details)
                    VALUES (?, ?, ?, ?)
                ''', (
                    datetime.now().isoformat(),
                    event_type,
                    str(path),
                    json.dumps(details)
                ))
                conn.commit()
        except Exception as e:
            self.logger.error(f"Error logging event: {str(e)}")

    def check_file_integrity(self, file_path):
        """Check file integrity against stored information"""
        try:
            current_info = self.get_file_info(file_path)
            if not current_info:
                return
            
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute(
                    'SELECT * FROM files WHERE path = ?',
                    (str(file_path),)
                )
                stored_info = cursor.fetchone()
                
                if stored_info:
                    # Compare current and stored information
                    changes = []
                    if current_info['hash'] != stored_info[1]:
                        changes.append('content')
                    if current_info['size'] != stored_info[2]:
                        changes.append('size')
                    if current_info['permissions'] != stored_info[3]:
                        changes.append('permissions')
                    if current_info['owner'] != stored_info[4]:
                        changes.append('owner')
                    
                    if changes:
                        self.log_event(
                            'modification',
                            str(file_path),
                            {'changes': changes}
                        )
                        self.event_queue.put({
                            'type': 'modification',
                            'path': str(file_path),
                            'changes': changes,
                            'timestamp': datetime.now().isoformat()
                        })
                
                # Update stored information
                self.store_file_info(current_info)
                
        except Exception as e:
            self.logger.error(
                f"Error checking integrity for {file_path}: {str(e)}"
            )

class FileSystemEventHandler(FileSystemEventHandler):
    def __init__(self, monitor):
        self.monitor = monitor

    def on_created(self, event):
        if not event.is_directory:
            self.monitor.event_queue.put({
                'type': 'creation',
                'path': event.src_path,
                'timestamp': datetime.now().isoformat()
            })
            self.monitor.log_event('creation', event.src_path, {})

    def on_modified(self, event):
        if not event.is_directory:
            self.monitor.check_file_integrity(event.src_path)

    def on_deleted(self, event):
        if not event.is_directory:
            self.monitor.event_queue.put({
                'type': 'deletion',
                'path': event.src_path,
                'timestamp': datetime.now().isoformat()
            })
            self.monitor.log_event('deletion', event.src_path, {})

    def on_moved(self, event):
        if not event.is_directory:
            self.monitor.event_queue.put({
                'type': 'move',
                'src_path': event.src_path,
                'dest_path': event.dest_path,
                'timestamp': datetime.now().isoformat()
            })
            self.monitor.log_event(
                'move',
                event.src_path,
                {'destination': event.dest_path}
            )

    def baseline_scan(self):
        """Perform baseline scan of monitored paths"""
        self.logger.info("Starting baseline scan...")
        
        for path in self.paths:
            if path.is_file():
                self.process_file(path)
            elif path.is_dir():
                self.process_directory(path)

    def process_file(self, file_path):
        """Process individual file during baseline scan"""
        try:
            file_info = self.get_file_info(file_path)
            if file_info:
                self.store_file_info(file_info)
        except Exception as e:
            self.logger.error(
                f"Error processing file {file_path}: {str(e)}"
            )

    def process_directory(self, directory):
        """Process directory during baseline scan"""
        try:
            for item in directory.rglob('*'):
                if item.is_file():
                    self.process_file(item)
        except Exception as e:
            self.logger.error(
                f"Error processing directory {directory}: {str(e)}"
            )

    def start_realtime_monitoring(self):
        """Start realtime file system monitoring"""
        self.logger.info("Starting realtime monitoring...")
        
        observer = Observer()
        handler = FileSystemEventHandler(self)
        
        for path in self.paths:
            if path.is_dir():
                observer.schedule(handler, str(path), recursive=True)
            else:
                observer.schedule(
                    handler,
                    str(path.parent),
                    recursive=False
                )
        
        observer.start()
        return observer

    def start_periodic_monitoring(self):
        """Start periodic file system monitoring"""
        self.logger.info(
            f"Starting periodic monitoring (interval: {self.interval}s)..."
        )
        
        while not self.stop_event.is_set():
            for path in self.paths:
                if path.is_file():
                    self.check_file_integrity(path)
                elif path.is_dir():
                    for file_path in path.rglob('*'):
                        if file_path.is_file():
                            self.check_file_integrity(file_path)
            
            time.sleep(self.interval)

    def generate_report(self, start_time, end_time=None):
        """Generate integrity monitoring report"""
        if not end_time:
            end_time = datetime.now().isoformat()
            
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                
                # Get all events in time range
                cursor.execute('''
                    SELECT * FROM events
                    WHERE timestamp BETWEEN ? AND ?
                    ORDER BY timestamp DESC
                ''', (start_time, end_time))
                events = cursor.fetchall()
                
                # Get files with modifications
                cursor.execute('''
                    SELECT * FROM files
                    WHERE last_modified BETWEEN ? AND ?
                ''', (start_time, end_time))
                modified_files = cursor.fetchall()
                
                report = {
                    'period': {
                        'start': start_time,
                        'end': end_time
                    },
                    'summary': {
                        'total_events': len(events),
                        'modified_files': len(modified_files),
                        'event_types': defaultdict(int)
                    },
                    'events': [],
                    'modified_files': []
                }
                
                # Process events
                for event in events:
                    event_info = {
                        'timestamp': event[0],
                        'type': event[1],
                        'path': event[2],
                        'details': json.loads(event[3])
                    }
                    report['summary']['event_types'][event[1]] += 1
                    report['events'].append(event_info)
                
                # Process modified files
                for file in modified_files:
                    file_info = {
                        'path': file[0],
                        'current_hash': file[1],
                        'size': file[2],
                        'permissions': file[3],
                        'owner': file[4],
                        'last_modified': file[5]
                    }
                    report['modified_files'].append(file_info)
                
                return report
                
        except Exception as e:
            self.logger.error(f"Error generating report: {str(e)}")
            return None

    def save_report(self, report, output_file):
        """Save monitoring report to file"""
        try:
            with open(output_file, 'w') as f:
                json.dump(report, f, indent=4)
            self.logger.info(f"Report saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving report: {str(e)}")

    def print_summary(self, report):
        """Print monitoring summary"""
        self.logger.info("\nFile Integrity Monitoring Summary:")
        self.logger.info(
            f"Period: {report['period']['start']} to {report['period']['end']}"
        )
        
        self.logger.info("\nEvent Summary:")
        self.logger.info(f"Total Events: {report['summary']['total_events']}")
        self.logger.info(f"Modified Files: {report['summary']['modified_files']}")
        
        self.logger.info("\nEvents by Type:")
        for event_type, count in report['summary']['event_types'].items():
            self.logger.info(f"{event_type}: {count}")
        
        if report['events']:
            self.logger.info("\nRecent Events:")
            for event in report['events'][:5]:  # Show last 5 events
                self.logger.info(
                    f"\nType: {event['type']}\n"
                    f"Path: {event['path']}\n"
                    f"Time: {event['timestamp']}"
                )
                if event.get('details'):
                    self.logger.info(f"Details: {event['details']}")

    def start_monitoring(self, output_file=None):
        """Start file integrity monitoring"""
        self.logger.info("Starting file integrity monitoring")
        self.logger.info(f"Mode: {self.mode}")
        self.logger.info("Monitored paths:")
        for path in self.paths:
            self.logger.info(f"- {path}")
        
        try:
            # Perform baseline scan
            self.baseline_scan()
            start_time = datetime.now().isoformat()
            
            # Start monitoring based on mode
            if self.mode == 'realtime':
                observer = self.start_realtime_monitoring()
                try:
                    while not self.stop_event.is_set():
                        time.sleep(1)
                except KeyboardInterrupt:
                    observer.stop()
                observer.join()
            else:
                self.start_periodic_monitoring()
                
        except KeyboardInterrupt:
            self.logger.info("Monitoring interrupted by user")
        except Exception as e:
            self.logger.error(f"Error during monitoring: {str(e)}")
        finally:
            # Generate and save report
            end_time = datetime.now().isoformat()
            report = self.generate_report(start_time, end_time)
            if report and output_file:
                self.save_report(report, output_file)
            if report:
                self.print_summary(report)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="File Integrity Monitor")
    parser.add_argument("paths", nargs='+',
                      help="Paths to monitor")
    parser.add_argument("--mode", choices=['realtime', 'periodic'],
                      default='realtime',
                      help="Monitoring mode")
    parser.add_argument("--interval", type=int, default=60,
                      help="Check interval for periodic mode (seconds)")
    parser.add_argument("--output",
                      help="Output file for monitoring report")
    args = parser.parse_args()

    monitor = FileIntegrityMonitor(args.paths, args.mode, args.interval)
    monitor.start_monitoring(args.output)
